// 2019.11.22 Kyoung Shin Park
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <thread>
#include <mutex>
#ifdef WIN32
#include <windows.h>
#else
#include <unistd.h>
#endif

#define MAX_NUM 5

struct BoundedBuffer
{
	std::string* buffer;
	int capacity;

	int head;
	int tail;
	int count;

	// The mutexes are managed by a std::unique_lock. It is a wrapper to manage a lock. This is necessary to be used with the condition variables. 
	std::mutex lock; // mutex lock
	// https://en.cppreference.com/w/cpp/thread/condition_variable/wait
	// condition_variable::wait(pred) equivalent to while (!pred()) { wait(lock) }
	// condition_variable::wait(pred) causes the current thread to block until the condition variable is notified or a spurious wakeup occurs, optionally looping until some predicate is satisfied. 
	// https://en.cppreference.com/w/cpp/thread/condition_variable/notify_one
	// condition_variable::notify_one() unblocks one of the waiting threads
	std::condition_variable not_full; // not_full condition_variable to block the calling thread until notified to resume
	std::condition_variable not_empty; // not_empty condition_variable to block the calling thread until notified to resume

	BoundedBuffer(int capacity) : capacity(capacity), head(0), tail(0), count(0) {
		buffer = new std::string[capacity];
	}

	~BoundedBuffer() {
		delete[] buffer;
	}

	void put(std::string data) {
		std::unique_lock<std::mutex> l(lock);
		// if buffer is full, wait until buffer is available (not_full: count <= capacity)
		// otherwise, no wait
		printf("put waiting......\n");
		not_full.wait(l, [this]() { return count != capacity; });
		printf("put finished waiting count %d != capacity %d.....\n", count, capacity);

		buffer[tail] = data;
		printf("\tcount=%d put buffer[%d]=%s\n", count, tail, buffer[tail].c_str());
		tail = (tail + 1) % capacity;
		++count;

		l.unlock();
		// notify not_empty
		not_empty.notify_one();
		printf("put notify not_empty......\n");
	}

	std::string take() {
		std::unique_lock<std::mutex> l(lock);
		// if buffer is empty, wait until buffer is available (not_empty: count >= 0)
		// otherwise, no wait
		printf("take waiting......\n");
		not_empty.wait(l, [this]() { return count != 0; });
		printf("take finished waiting count %d != 0.....\n", count);

		std::string result = buffer[head];
		printf("\t\tcount=%d take buffer[%d]=%s\n", count, head, result.c_str());
		head = (head + 1) % capacity;
		--count;

		l.unlock();
		// notify not_full_cv
		not_full.notify_one();

		return result;
	}
};

void writer(int id, BoundedBuffer& buffer) {
	std::string item;
	for (int i = 0; i < 2*MAX_NUM; ++i) {
		item = "Writer " + std::to_string(id) + " produced " + std::to_string(i);
		printf("%s\n", item.c_str());
		buffer.put(item);
		std::this_thread::sleep_for(std::chrono::milliseconds(1000));
	}
}

void reader(int id, BoundedBuffer& buffer) {
	while (true) {
		std::string item = buffer.take();
		printf("\t\t-------Reader %d consumed %s\n", id, item.c_str());
	}
}

int main(int argc, char** argv)
{
	BoundedBuffer buffer(MAX_NUM);

	std::thread th[5];	
	th[0] = std::thread(writer, 0, std::ref(buffer));
	th[1] = std::thread(writer, 1, std::ref(buffer));
	th[2] = std::thread(writer, 2, std::ref(buffer));
#ifdef WIN32
	Sleep(5000); // window Sleep
#else
	usleep(5000); // linux usleep
#endif
	th[3] = std::thread(reader, 0, std::ref(buffer));
	th[4] = std::thread(reader, 1, std::ref(buffer));

	th[0].join();
	th[1].join();
	th[2].join();
	th[3].join();
	th[4].join();

	return 0;
}